// SPDX-License-Identifier: GPL-3.0

pragma solidity 0.6.8;
pragma experimental ABIEncoderV2;

import { SafeMath } from "@openzeppelin/contracts/math/SafeMath.sol";
import { IERC721, IERC165 } from "@openzeppelin/contracts/token/ERC721/IERC721.sol";
import { ReentrancyGuard } from "@openzeppelin/contracts/utils/ReentrancyGuard.sol";
import {IERC20} from "@openzeppelin/contracts/token/ERC20/IERC20.sol";
import { SafeERC20 } from "@openzeppelin/contracts/token/ERC20/SafeERC20.sol";
import {Counters} from "@openzeppelin/contracts/utils/Counters.sol";
import { IItemExchange, Decimal } from "./interfaces/IItemExchange.sol";
import { IItem } from "./interfaces/IItem.sol";
import { IItemListing } from "./interfaces/IItemListing.sol"; 

interface IWETH {
    function deposit() external payable;
    function withdraw(uint wad) external;

    function transfer(address to, uint256 value) external returns (bool);
}

interface IItemExtended is IItem {
    function itemExchangeContract() external returns(address);
}
 
contract ItemListing is IItemListing, ReentrancyGuard {
    using SafeMath for uint256;
    using SafeERC20 for IERC20;
    using Counters for Counters.Counter; 
    uint256 public timeBuffer; 
    uint8 public minBidIncrementPercentage; 
    address public motif; 
    address public wethAddress; 
    mapping(uint256 => IItemListing.Listing) public listings; 
    
    bytes4 constant interfaceId = 0x80ac58cd;  

    Counters.Counter private _listingIdTracker;

    modifier listingExists(uint256 listingId) {
        require(_exists(listingId), "Listing doesn't exist");
        _;
    }

    /*
     * Constructor
     */
    constructor(address _motif, address _weth) public {
        require(
            IERC165(_motif).supportsInterface(interfaceId),
            "Doesn't support NFT interface"
        );
        motif = _motif;
        wethAddress = _weth;
        timeBuffer = 15 * 60; // extend 15 minutes after every bid made in last 15 minutes
        minBidIncrementPercentage = 5; // 5%
    }

    /**
     * @notice Create an listing.
     * @dev Store the listing details in the listings mapping and emit an ListingCreated event.
     * If there is no curator, or if the curator is the listing creator, automatically approve the listing.
     */
    function createListing(
        uint256 tokenId,
        address tokenContract,
        uint256 duration,
        uint256 reservePrice,
        address payable curator,
        uint8 curatorFeePercentage,
        address listingCurrency
    ) public override nonReentrant returns (uint256) {
        require(
            IERC165(tokenContract).supportsInterface(interfaceId),
            "tokenContract does not support ERC721 interface"
        );
        require(curatorFeePercentage < 100, "curatorFeePercentage must be less than 100");
        address tokenOwner = IERC721(tokenContract).ownerOf(tokenId);
        require(msg.sender == IERC721(tokenContract).getApproved(tokenId) || msg.sender == tokenOwner, "Caller must be approved or owner for token id");
        uint256 listingId = _listingIdTracker.current();

        listings[listingId] = Listing({
            tokenId: tokenId,
            tokenContract: tokenContract,
            approved: false,
            amount: 0,
            duration: duration,
            firstBidTime: 0,
            reservePrice: reservePrice,
            curatorFeePercentage: curatorFeePercentage,
            tokenOwner: tokenOwner,
            bidder: address(0),
            curator: curator,
            listingCurrency: listingCurrency
        });

        IERC721(tokenContract).transferFrom(tokenOwner, address(this), tokenId);

        _listingIdTracker.increment();

        emit ListingCreated(listingId, tokenId, tokenContract, duration, reservePrice, tokenOwner, curator, curatorFeePercentage, listingCurrency);


        if(listings[listingId].curator == address(0) || curator == tokenOwner) {
            _approveListing(listingId, true);
        }

        return listingId;
    }

    /**
     * @notice Approve an listing, opening up the listing for bids.
     * @dev Only callable by the curator. Cannot be called if the listing has already started.
     */
    function setListingApproval(uint256 listingId, bool approved) external override listingExists(listingId) {
        require(msg.sender == listings[listingId].curator, "Must be listing curator");
        require(listings[listingId].firstBidTime == 0, "Listing has already started");
        _approveListing(listingId, approved);
    }

    function setListingReservePrice(uint256 listingId, uint256 reservePrice) external override listingExists(listingId) {
        require(msg.sender == listings[listingId].curator || msg.sender == listings[listingId].tokenOwner, "Must be listing curator or token owner");
        require(listings[listingId].firstBidTime == 0, "Listing has already started");

        listings[listingId].reservePrice = reservePrice;

        emit ListingReservePriceUpdated(listingId, listings[listingId].tokenId, listings[listingId].tokenContract, reservePrice);
    }

    /**
     * @notice Create a bid on a token, with a given amount.
     * @dev If provided a valid bid, transfers the provided amount to this contract.
     * If the listing is run in native ETH, the ETH is wrapped so it can be identically to other
     * listing currencies in this contract.
     */
    function createBid(uint256 listingId, uint256 amount)
    external
    override
    payable
    listingExists(listingId)
    nonReentrant
    {
        address payable lastBidder = listings[listingId].bidder;
        require(listings[listingId].approved, "Listing must be approved by curator");
        require(
            listings[listingId].firstBidTime == 0 ||
            block.timestamp <
            listings[listingId].firstBidTime.add(listings[listingId].duration),
            "Listing expired"
        );
        require(
            amount >= listings[listingId].reservePrice,
                "Must send at least reservePrice"
        );
        require(
            amount >= listings[listingId].amount.add(
                listings[listingId].amount.mul(minBidIncrementPercentage).div(100)
            ),
            "Must send more than last bid by minBidIncrementPercentage amount"
        );

        // For Motif Protocol, ensure that the bid is valid for the current bidShare configuration
        if(listings[listingId].tokenContract == motif) {
            require(
                IItemExchange(IItemExtended(motif).itemExchangeContract()).isValidBid(
                    listings[listingId].tokenId,
                    amount
                ),
                "Bid invalid for share splitting"
            );
        }

        // If this is the first valid bid, we should set the starting time now.
        // If it's not, then we should refund the last bidder
        if(listings[listingId].firstBidTime == 0) {
            listings[listingId].firstBidTime = block.timestamp;
        } else if(lastBidder != address(0)) {
            _handleOutgoingBid(lastBidder, listings[listingId].amount, listings[listingId].listingCurrency);
        }

        _handleIncomingBid(amount, listings[listingId].listingCurrency);

        listings[listingId].amount = amount;
        listings[listingId].bidder = msg.sender;


        bool extended = false;
        // at this point we know that the timestamp is less than start + duration (since the listing would be over, otherwise)
        // we want to know by how much the timestamp is less than start + duration
        // if the difference is less than the timeBuffer, increase the duration by the timeBuffer
        if (
            listings[listingId].firstBidTime.add(listings[listingId].duration).sub(
                block.timestamp
            ) < timeBuffer
        ) {
            // Playing code golf for gas optimization:
            // uint256 expectedEnd = listings[listingId].firstBidTime.add(listings[listingId].duration);
            // uint256 timeRemaining = expectedEnd.sub(block.timestamp);
            // uint256 timeToAdd = timeBuffer.sub(timeRemaining);
            // uint256 newDuration = listings[listingId].duration.add(timeToAdd);
            uint256 oldDuration = listings[listingId].duration;
            listings[listingId].duration =
                oldDuration.add(timeBuffer.sub(listings[listingId].firstBidTime.add(oldDuration).sub(block.timestamp)));
            extended = true;
        }

        emit ListingBid(
            listingId,
            listings[listingId].tokenId,
            listings[listingId].tokenContract,
            msg.sender,
            amount,
            lastBidder == address(0), // firstBid boolean
            extended
        );

        if (extended) {
            emit ListingDurationExtended(
                listingId,
                listings[listingId].tokenId,
                listings[listingId].tokenContract,
                listings[listingId].duration
            );
        }
    }

    /**
     * @notice End an listing, finalizing the bid on Motif if applicable and paying out the respective parties.
     * @dev If for some reason the listing cannot be finalized (invalid token recipient, for example),
     * The listing is reset and the NFT is transferred back to the listing creator.
     */
    function endListing(uint256 listingId) external override listingExists(listingId) nonReentrant {
        require(
            uint256(listings[listingId].firstBidTime) != 0,
            "Listing hasn't begun"
        );
        require(
            block.timestamp >=
            listings[listingId].firstBidTime.add(listings[listingId].duration),
            "Listing hasn't completed"
        );

        address currency = listings[listingId].listingCurrency == address(0) ? wethAddress : listings[listingId].listingCurrency;
        uint256 curatorFee = 0;

        uint256 tokenOwnerProfit = listings[listingId].amount;

        if(listings[listingId].tokenContract == motif) {
            // If the listing is running on motif, settle it on the protocol
            (bool success, uint256 remainingProfit) = _handleMotifListingSettlement(listingId);
            tokenOwnerProfit = remainingProfit;
            if(success != true) {
                _handleOutgoingBid(listings[listingId].bidder, listings[listingId].amount, listings[listingId].listingCurrency);
                _cancelListing(listingId);
                return;
            }
        } else {
            // Otherwise, transfer the token to the winner and pay out the participants below
            try IERC721(listings[listingId].tokenContract).safeTransferFrom(address(this), listings[listingId].bidder, listings[listingId].tokenId) {} catch {
                _handleOutgoingBid(listings[listingId].bidder, listings[listingId].amount, listings[listingId].listingCurrency);
                _cancelListing(listingId);
                return;
            }
        }


        if(listings[listingId].curator != address(0)) {
            curatorFee = tokenOwnerProfit.mul(listings[listingId].curatorFeePercentage).div(100);
            tokenOwnerProfit = tokenOwnerProfit.sub(curatorFee);
            _handleOutgoingBid(listings[listingId].curator, curatorFee, listings[listingId].listingCurrency);
        }
        _handleOutgoingBid(listings[listingId].tokenOwner, tokenOwnerProfit, listings[listingId].listingCurrency);

        emit ListingEnded(
            listingId,
            listings[listingId].tokenId,
            listings[listingId].tokenContract,
            listings[listingId].tokenOwner,
            listings[listingId].curator,
            listings[listingId].bidder,
            tokenOwnerProfit,
            curatorFee,
            currency
        );
        delete listings[listingId];
    }

    /**
     * @notice Cancel an listing.
     * @dev Transfers the NFT back to the listing creator and emits an ListingCanceled event
     */
    function cancelListing(uint256 listingId) external override nonReentrant listingExists(listingId) {
        require(
            listings[listingId].tokenOwner == msg.sender || listings[listingId].curator == msg.sender,
            "Can only be called by listing creator or curator"
        );
        require(
            uint256(listings[listingId].firstBidTime) == 0,
            "Can't cancel an listing once it's begun"
        );
        _cancelListing(listingId);
    }

    /**
     * @dev Given an amount and a currency, transfer the currency to this contract.
     * If the currency is ETH (0x0), attempt to wrap the amount as WETH
     */
    function _handleIncomingBid(uint256 amount, address currency) internal {
        // If this is an ETH bid, ensure they sent enough and convert it to WETH under the hood
        if(currency == address(0)) {
            require(msg.value == amount, "Sent ETH Value does not match specified bid amount");
            IWETH(wethAddress).deposit{value: amount}();
        } else {
            // We must check the balance that was actually transferred to the listing,
            // as some tokens impose a transfer fee and would not actually transfer the
            // full amount to the itemExchange, resulting in potentally locked funds
            IERC20 token = IERC20(currency);
            uint256 beforeBalance = token.balanceOf(address(this));
            token.safeTransferFrom(msg.sender, address(this), amount);
            uint256 afterBalance = token.balanceOf(address(this));
            require(beforeBalance.add(amount) == afterBalance, "Token transfer call did not transfer expected amount");
        }
    }

    function _handleOutgoingBid(address to, uint256 amount, address currency) internal {
        // If the listing is in ETH, unwrap it from its underlying WETH and try to send it to the recipient.
        if(currency == address(0)) {
            IWETH(wethAddress).withdraw(amount);

            // If the ETH transfer fails (sigh), rewrap the ETH and try send it as WETH.
            if(!_safeTransferETH(to, amount)) {
                IWETH(wethAddress).deposit{value: amount}();
                IERC20(wethAddress).safeTransfer(to, amount);
            }
        } else {
            IERC20(currency).safeTransfer(to, amount);
        }
    }

    function _safeTransferETH(address to, uint256 value) internal returns (bool) {
        (bool success, ) = to.call{value: value}(new bytes(0));
        return success;
    }

    function _cancelListing(uint256 listingId) internal {
        address tokenOwner = listings[listingId].tokenOwner;
        IERC721(listings[listingId].tokenContract).safeTransferFrom(address(this), tokenOwner, listings[listingId].tokenId);

        emit ListingCanceled(listingId, listings[listingId].tokenId, listings[listingId].tokenContract, tokenOwner);
        delete listings[listingId];
    }

    function _approveListing(uint256 listingId, bool approved) internal {
        listings[listingId].approved = approved;
        emit ListingApprovalUpdated(listingId, listings[listingId].tokenId, listings[listingId].tokenContract, approved);
    }

    function _exists(uint256 listingId) internal view returns(bool) {
        return listings[listingId].tokenOwner != address(0);
    }

    function _handleMotifListingSettlement(uint256 listingId) internal returns (bool, uint256) {
        address currency = listings[listingId].listingCurrency == address(0) ? wethAddress : listings[listingId].listingCurrency;

        IItemExchange.Bid memory bid = IItemExchange.Bid({
            amount: listings[listingId].amount,
            currency: currency,
            bidder: address(this),
            recipient: listings[listingId].bidder,
            sellOnShare: Decimal.D256(0)
        });

        IERC20(currency).approve(IItemExtended(motif).itemExchangeContract(), bid.amount);
        IItem(motif).setBid(listings[listingId].tokenId, bid);
        uint256 beforeBalance = IERC20(currency).balanceOf(address(this));
        try IItem(motif).acceptBid(listings[listingId].tokenId, bid) {} catch {
            // If the underlying NFT transfer here fails, we should cancel the listing and refund the winner
            IItemExtended(motif).removeBid(listings[listingId].tokenId);
            return (false, 0);
        }
        uint256 afterBalance = IERC20(currency).balanceOf(address(this));

        // We have to calculate the amount to send to the token owner here in case there was a
        // sell-on share on the token
        return (true, afterBalance.sub(beforeBalance));
    }

    // TODO: consider reverting if the message sender is not WETH
    receive() external payable {}
    fallback() external payable {}
}