// SPDX-License-Identifier: GPL-3.0

pragma solidity 0.6.8;
pragma experimental ABIEncoderV2;

/**
 * @title Interface for Item Listings
 */
interface IItemListing {
    struct Listing {
        // ID for the ERC721 token
        uint256 tokenId;
        // Address for the ERC721 contract
        address tokenContract;
        // Whether or not the listing curator has approved the listing to start
        bool approved;
        // The current highest bid amount
        uint256 amount;
        // The length of time to run the listing for, after the first bid was made
        uint256 duration;
        // The time of the first bid
        uint256 firstBidTime;
        // The minimum price of the first bid
        uint256 reservePrice;
        // The sale percentage to send to the curator
        uint8 curatorFeePercentage;
        // The address that should receive the funds once the NFT is sold.
        address tokenOwner;
        // The address of the current highest bid
        address payable bidder;
        // The address of the listing's curator.
        // The curator can reject or approve an listing
        address payable curator;
        // The address of the ERC-20 currency to run the listing with.
        // If set to 0x0, the listing will be run in ETH
        address listingCurrency;
    }

    event ListingCreated(
        uint256 indexed listingId,
        uint256 indexed tokenId,
        address indexed tokenContract,
        uint256 duration,
        uint256 reservePrice,
        address tokenOwner,
        address curator,
        uint8 curatorFeePercentage,
        address listingCurrency
    );

    event ListingApprovalUpdated(
        uint256 indexed listingId,
        uint256 indexed tokenId,
        address indexed tokenContract,
        bool approved
    );

    event ListingReservePriceUpdated(
        uint256 indexed listingId,
        uint256 indexed tokenId,
        address indexed tokenContract,
        uint256 reservePrice
    );

    event ListingBid(
        uint256 indexed listingId,
        uint256 indexed tokenId,
        address indexed tokenContract,
        address sender,
        uint256 value,
        bool firstBid,
        bool extended
    );

    event ListingDurationExtended(
        uint256 indexed listingId,
        uint256 indexed tokenId,
        address indexed tokenContract,
        uint256 duration
    );

    event ListingEnded(
        uint256 indexed listingId,
        uint256 indexed tokenId,
        address indexed tokenContract,
        address tokenOwner,
        address curator,
        address winner,
        uint256 amount,
        uint256 curatorFee,
        address listingCurrency
    );

    event ListingCanceled(
        uint256 indexed listingId,
        uint256 indexed tokenId,
        address indexed tokenContract,
        address tokenOwner
    );

    function createListing(
        uint256 tokenId,
        address tokenContract,
        uint256 duration,
        uint256 reservePrice,
        address payable curator,
        uint8 curatorFeePercentages,
        address listingCurrency
    ) external returns (uint256);

    function setListingApproval(uint256 listingId, bool approved) external;

    function setListingReservePrice(uint256 listingId, uint256 reservePrice) external;

    function createBid(uint256 listingId, uint256 amount) external payable;

    function endListing(uint256 listingId) external;

    function cancelListing(uint256 listingId) external;
}